# Copyright (C) 2023  Nexedi SA and Contributors.
#                     Started by Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.


import nxdbom
import pytest

import os
from os.path import dirname, exists


@pytest.mark.parametrize('url,nameok,verok', [
    ('http://www.python.org/ftp/python/2.7.18/Python-2.7.18.tar.xz',  'Python',  '2.7.18'),
    ('http://www.ijg.org/files/jpegsrc.v9d.tar.gz',                   'jpegsrc', '9d'),
    ('https://github.com/nghttp2/nghttp2/archive/v1.40.0.tar.gz',     'nghttp2', '1.40.0'),
    ('https://golang.org/dl/go1.18.9.src',                            'go',      '1.18.9'),
    ('https://github.com/tesseract-ocr/tesseract/archive/refs/tags/4.1.1.tar.gz',  'tesseract', '4.1.1'),
    ('https://raw.githubusercontent.com/zuphilip/ocropy-models/master/en-default.pyrnn.gz',  'ocropy-models', 'master'),
    ('https://sourceforge.net/projects/swig/files/swig/swig-3.0.12/swig-3.0.12.tar.gz/download',  'swig', '3.0.12'),
    ('https://git.savannah.gnu.org/gitweb/?p=config.git;a=snapshot;h=5e531d39;sf=tgz',  'config', '5e531d39'),
    ('https://lab.nexedi.com/nexedi/wendelin.core.git',  'wendelin.core', None),
    ('/ROOT/develop-eggs/mysqlclient-1.3.12-py2.7-linux-x86_64.egg',  'mysqlclient', '1.3.12'),
    ('https://osdn.net/frs/redir.php?f=ipafonts%2F57330%2FIPAexfont00201.zip', 'IPAexfont', '00201'),
    ('https://osdn.net/frs/redir.php?f=ipafonts%2F51868%2FIPAfont00303.zip',   'IPAfont',   '00303'),
    ('https://osdn.net/frs/redir.php?f=tsukurimashou%2F56948%2Focr-0.2.zip',   'ocr',       '0.2'),
    ('http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz', 'GeoLite2-Country', None),
    ('http://downloadarchive.documentfoundation.org/libreoffice/old/5.2.4.2/rpm/x86_64/LibreOffice_5.2.4.2_Linux_x86-64_rpm.tar.gz',  'LibreOffice', '5.2.4.2'),
    ('http://www.cups.org/software/1.7.4/cups-1.7.4-source.tar.bz2',  'cups', '1.7.4'),
])
def test_namever(url, nameok, verok):
    assert nxdbom.namever(url) == (nameok, verok)

@pytest.mark.parametrize('url', [
    'xxx.conf.in',
    'xxx.cfg',
    'xxx.cfg.in',
    'xxx.cnf.in',
    'xxx.cfg.jinja2',
    'xxx.cfg.jinja2.in',
    'xxx.jinja2.cfg',
    'xxx.jinja2.library',
    'xxx.asn',
    'xxx/ltelogs.jinja2.sh',
    'xxx/templates/wrapper.in',
    'xxx/logrotate_entry.in',
    'xxx/matplotlibrc.in',
    'xxx/goenv.sh.in',
    'xxx/promise/yyy',
])
def test_isconf(url):
    assert nxdbom.isconf(url) == True


# ---- BOM software ----

testv = []  # of (build, bomok)
def case1(build, bomok):
    testv.append((build, bomok))

case1("""\
[ncurses]
recipe = slapos.recipe.cmmi
url = http://ftp.gnu.org/gnu/ncurses/ncurses-6.2.tar.gz
""", """\
ncurses                      6.2        http://ftp.gnu.org/gnu/ncurses/ncurses-6.2.tar.gz
""")

case1("""\
[neoppod-repository]
recipe = slapos.recipe.build:gitclone
repository = https://lab.nexedi.com/nexedi/neoppod.git
""", """
>>> gits:
neoppod                      HEAD       https://lab.nexedi.com/nexedi/neoppod.git
""")

case1("""\
[ocropy-eng-traineddata]
recipe = slapos.recipe.build:download
url = https://raw.githubusercontent.com/zuphilip/ocropy-models/master/en-default.pyrnn.gz
""", """\
ocropy-models                master     https://raw.githubusercontent.com/zuphilip/ocropy-models/master/en-default.pyrnn.gz
""")

case1("""\
[scons]
recipe = slapos.recipe.build:download-unpacked
url = https://prdownloads.sourceforge.net/scons/scons-local-2.3.0.tar.gz
""", """\
scons-local                  2.3.0      https://prdownloads.sourceforge.net/scons/scons-local-2.3.0.tar.gz
""")

case1("""\
[ipaex-fonts]
recipe = slapos.recipe.build:download-unpacked
url = https://osdn.net/frs/redir.php?f=ipafonts%2F57330%2FIPAexfont00201.zip
""", """\
IPAexfont                    00201      https://osdn.net/frs/redir.php?f=ipafonts%2F57330%2FIPAexfont00201.zip
""")

# download from local filesystem - ignore if that file comes from the profile itself
case1("""\
[mariadb-resiliency-after-import-script]
recipe = slapos.recipe.build:download
url = /BASE/stack/erp5/instance-mariadb-resiliency-after-import-script.sh.in
_profile_base_location_ = /BASE
""", '')

# slapos.build.:download with hash in the url
case1("""\
[geolite2-country]
recipe = slapos.recipe.build:download-unpacked
url = http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz#dc6224c648350d90f344a0c5c3ca5474
""", """\
GeoLite2-Country             dc6224c6   http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz
""")

for x in ('gcc', 'python', 'ZODB', 'ZEO', 'tempstorage'):
    case1("""\
[%s]
recipe = slapos.recipe.build
""" % x, '')    # empty


case1("""\
[template-logrotate-base]
recipe = slapos.recipe.template:jinja2
url = /srv/slapgrid/slappart47/srv/project/slapos/stack/logrotate/instance-logrotate-base.cfg.in
""", '')    # config ignored

case1("""\
[cfg-environment]
recipe = collective.recipe.template
input = inline: zzz
""", '')    # inline ignore


case1("""\
[neoppod-develop]
recipe = zc.recipe.egg:develop
__buildout_installed__ = /ROOT/develop-eggs/neoppod.egg-link
""", '')    # .egg-link ignored

case1("""\
[python-mysqlclient]
recipe = zc.recipe.egg:custom
__buildout_installed__ = /ROOT/develop-eggs/mysqlclient-1.3.12-py2.7-linux-x86_64.egg
""", """
>>> eggs:
mysqlclient                  1.3.12     https://pypi.org/project/mysqlclient/1.3.12/
""")

case1("""\
[slapos-toolbox-dependencies]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = lxml
        pycurl
        Mako
-- /ROOT/develop-eggs/lxml-4.9.1-py2.7-linux-x86_64.egg/x --
-- /ROOT/develop-eggs/pycurl-7.43.0-py2.7-linux-x86_64.egg/x --
-- /ROOT/eggs/Mako-1.1.4-py2.7.egg/x --
""", """
>>> eggs:
Mako                         1.1.4      https://pypi.org/project/Mako/1.1.4/
lxml                         4.9.1      https://pypi.org/project/lxml/4.9.1/
pycurl                       7.43.0     https://pypi.org/project/pycurl/7.43.0/
""")

case1("""\
[ZODB5]
recipe = zc.recipe.egg:eggs
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = ZODB
      BTrees
-- /ROOT/eggs/ZODB-5.8.0-py2.7.egg/x --
-- /ROOT/develop-eggs/BTrees-4.11.3-py2.7-linux-x86_64.egg/x --
""", """
>>> eggs:
BTrees                       4.11.3     https://pypi.org/project/BTrees/4.11.3/
ZODB                         5.8.0      https://pypi.org/project/ZODB/5.8.0/
""")

case1("""\
[erp5-python-interpreter]
recipe = zc.recipe.egg:scripts
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = pygolang
        neoppod[admin, ctl, master]
-- /ROOT/develop-eggs/pygolang-0.1-py2.7-linux-x86_64.egg/x --
-- /ROOT/develop-eggs/neoppod.egg-link --
""", """
>>> eggs:
pygolang                     0.1        https://pypi.org/project/pygolang/0.1/
""")

# no `eggs =`
case1("""\
[selenium]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
-- /ROOT/eggs/selenium-3.141.0-py2.7.egg/x --
""", """
>>> eggs:
selenium                     3.141.0    https://pypi.org/project/selenium/3.141.0/
""")

# msgpack as egg and as c library at the same time
case1("""
[messagepack]
recipe = slapos.recipe.cmmi
url = http://downloads.sourceforge.net/project/msgpack/msgpack/cpp/msgpack-0.5.4.tar.gz

[msgpack-python]
recipe = zc.recipe.egg:custom
__buildout_installed__ = /ROOT/develop-eggs/msgpack-0.6.2-py2.7-linux-x86_64.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
egg = msgpack
""", """\
msgpack                      0.5.4      http://downloads.sourceforge.net/project/msgpack/msgpack/cpp/msgpack-0.5.4.tar.gz

>>> eggs:
msgpack                      0.6.2      https://pypi.org/project/msgpack/0.6.2/
""")

# %(__buildout_space_...) in egg can be read
case1("""\
[zzz]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs =
    aaa
    bbb%(__buildout_space_n__)s
-- /ROOT/eggs/aaa-1.2.3.egg/x --
-- /ROOT/eggs/bbb-5.6.7.egg/x --
""", """
>>> eggs:
aaa                          1.2.3      https://pypi.org/project/aaa/1.2.3/
bbb                          5.6.7      https://pypi.org/project/bbb/5.6.7/
""")

# +slapospatchedXXX is removed from egg URL
case1("""\
[astroid]
recipe = zc.recipe.egg:custom
__buildout_installed__ = /ROOT/develop-eggs/astroid-1.3.8+slapospatched001-py2.7.egg
""", """
>>> eggs:
astroid                      1.3.8+slapospatched001 https://pypi.org/project/astroid/1.3.8/
""")

# nxd or slaposXXX in egg version -> nexedi.org
case1("""\
[testrunner]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = SOAPpy
       zc.recipe.egg
       zc.buildout
-- /ROOT/eggs/SOAPpy-0.12.0nxd001-py2.7.egg/x --
-- /ROOT/eggs/zc.buildout-2.7.1+slapos019-py3.7.egg/x --
-- /ROOT/eggs/zc.recipe.egg-2.0.3+slapos003-py2.7.egg/x --
""", """
>>> eggs:
SOAPpy                       0.12.0nxd001 http://www.nexedi.org/static/packages/source/SOAPpy-0.12.0nxd001.tar.gz
zc.buildout                  2.7.1+slapos019 http://www.nexedi.org/static/packages/source/slapos.buildout/zc.buildout-2.7.1+slapos019.tar.gz
zc.recipe.egg                2.0.3+slapos003 http://www.nexedi.org/static/packages/source/zc.recipe.egg-2.0.3+slapos003.tar.gz
""")

# %20 in URL
case1("""\
[zabbix-agent]
recipe = slapos.recipe.cmmi
url = http://downloads.sourceforge.net/project/zabbix/ZABBIX%20Latest%20Stable/2.0.10/zabbix-2.0.10.tar.gz
""", """\
zabbix                       2.0.10     http://downloads.sourceforge.net/project/zabbix/ZABBIX%20Latest%20Stable/2.0.10/zabbix-2.0.10.tar.gz
""")

# plone.recipe.command is ignored
case1("""\
[gowork.dir]
recipe = plone.recipe.command
""", '')

# readline 8 and 5 used simultaneously    ->  older one is presented as readline5
# libpng   12 and 16 used simultaneously  ->  ----//----                libpng12
case1("""\
[readline]
recipe = slapos.recipe.cmmi
url = http://ftp.gnu.org/gnu/readline/readline-8.1.tar.gz
[readline5]
recipe = slapos.recipe.cmmi
url = http://ftp.gnu.org/gnu/readline/readline-5.2.tar.gz

[libpng]
recipe = slapos.recipe.cmmi
url = http://download.sourceforge.net/libpng/libpng-1.6.37.tar.xz
[libpng12]
recipe = slapos.recipe.cmmi
url = http://download.sourceforge.net/libpng/libpng-1.2.59.tar.xz

""", """\
libpng                       1.6.37     http://download.sourceforge.net/libpng/libpng-1.6.37.tar.xz
libpng12                     1.2.59     http://download.sourceforge.net/libpng/libpng-1.2.59.tar.xz
readline                     8.1        http://ftp.gnu.org/gnu/readline/readline-8.1.tar.gz
readline5                    5.2        http://ftp.gnu.org/gnu/readline/readline-5.2.tar.gz
""")

# egg directoty on the filesystem might have different case and _ instead of -
case1("""\
[cython-zstd]
recipe = zc.recipe.egg:custom
egg = cython-zstd
__buildout_installed__ = /ROOT/develop-eggs/cython_zstd-0.2-py2.7-linux-x86_64.egg

[cython]
recipe = zc.recipe.egg:custom
egg = cython
__buildout_installed__ = /Root/develop-eggs/Cython-0.29.24-py2.7-linux-x86_64.egg

[neoppod]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = cython-zstd
       cython
-- /ROOT/develop-eggs/cython_zstd-0.2-py2.7-linux-x86_64.egg/x --
-- /ROOT/develop-eggs/Cython-0.29.24-py2.7-linux-x86_64.egg/x --
""", """
>>> eggs:
Cython                       0.29.24    https://pypi.org/project/Cython/0.29.24/
cython_zstd                  0.2        https://pypi.org/project/cython_zstd/0.2/
""")

# py3 eggs
case1("""\
[xxx]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = aaa
       bbb
       ccc
-- /ROOT/eggs/aaa-1.2-py3.7.egg/x --
-- /ROOT/eggs/bbb-3.4-py3.8.egg/x --
-- /ROOT/develop-eggs/ccc-5.6.7-py3.9-linux-x86_64.egg/x --
""", """
>>> eggs:
aaa                          1.2        https://pypi.org/project/aaa/1.2/
bbb                          3.4        https://pypi.org/project/bbb/3.4/
ccc                          5.6.7      https://pypi.org/project/ccc/5.6.7/
""")

# libreoffice
case1("""\
[libreoffice-bin]
recipe = slapos.recipe.build
url = http://downloadarchive.documentfoundation.org/libreoffice/old/5.2.4.2/rpm/x86_64/LibreOffice_5.2.4.2_Linux_x86-64_rpm.tar.gz
""", """\
LibreOffice                  5.2.4.2    http://downloadarchive.documentfoundation.org/libreoffice/old/5.2.4.2/rpm/x86_64/LibreOffice_5.2.4.2_Linux_x86-64_rpm.tar.gz
""")


@pytest.mark.parametrize('build,bomok', testv)
def test_bom_software(tmpdir, build, bomok):
    tmpdir = str(tmpdir)
    build = '-- /ROOT/.installed.cfg --\n' + build
    build = build.replace('/ROOT', tmpdir)
    build = build.replace('/BASE', tmpdir+'/base')
    ar = txtar_parse(build)
    assert ar.comment == ''
    for f, data in ar.files.items():
        assert f.startswith(tmpdir)
        os.makedirs(dirname(f), exist_ok=True)
        with open(f, 'w') as _:
            _.write(data)

    bom = nxdbom.bom_software(tmpdir)
    assert nxdbom.fmt_bom(bom) == bomok


# loading non-existing .installed.cfg -> error
def test_bom_software_eexist():
    ne = '/nonexisting'
    assert not exists(ne)
    with pytest.raises(RuntimeError, match="Cannot load '%s/.installed.cfg'" % ne):
        nxdbom.bom_software(ne)


# ---- txtar ----

# txtar_* provide support for archives in txtar format
# https://pkg.go.dev/golang.org/x/tools/txtar#hdr-Txtar_format
class txtar_Archive:
    # .comment  str
    # .files    {}  name -> text
    pass

def txtar_parse(text): # -> txtar_Archive
    comment = ''
    files = {}
    current_file = None # current file | None for comments
    current_text = ''   # accumulator for text of current file
    def flush(next_file):
        nonlocal comment, current_file, current_text
        if current_file is None:
            comment = current_text
        else:
            files[current_file] = current_text

        current_text = ''
        current_file = next_file

    for l in text.splitlines(keepends=True):
        if not l.endswith('\n'):
            l += '\n'   # missing trailing newline on the final line

        if l.startswith('-- ') and l.rstrip().endswith(' --'):
            _ = l.rstrip()
            _ = _.removeprefix('-- ')
            _ = _.removesuffix(' --')
            next_file = _.strip()
            flush(next_file)
            continue

        current_text += l

    flush(None)
    ar = txtar_Archive()
    ar.comment = comment
    ar.files   = files
    return ar
